/*
	romtap - roms to ZX Spectrum .tap file
	Copyright (C) 2010 Edward Cree

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#define min(a,b)	((a)<(b)?(a):(b))

int blk2tap(FILE *out, char blk[16384], char * filename);

int main(int argc, char *argv[])
{
	if(argc>2)
	{
		fprintf(stderr, "M ROMTAP no good, too many arguments\n");
		return(255);
	}
	else
	{
		char *rom, *tap;
		FILE *romf, *tapf;
		if(argc==2)
		{
			rom=strdup(argv[1]);
			tap=strdup(rom);
			char * dot=strchr(tap, '.');
			if(!dot)
				dot=tap+strlen(tap);
			while(1)
			{
				char *next=strchr(dot+1, '.');
				if(next)
					dot=next;
				else
					break;
			}
			*dot=0;
			char *ntap=(char *)malloc(strlen(tap)+4);
			sprintf(ntap, "%s.tap", tap);
			free(tap);
			tap=ntap;
			romf=fopen(rom, "rb");
			tapf=fopen(tap, "wb");
		}
		else
		{
			rom="<stdin>";
			tap="<stdout>";
			romf=stdin;
			tapf=stdout;
		}
		if(!romf)
		{
			fprintf(stderr, "M ROMTAP no good, failed to open input file %s\n", rom);
			perror("fopen");
			return(1);
		}
		char blk[16384];
		int i;
		for(i=0;i<16384;i++)
		{
			int c=fgetc(romf);
			if(c==EOF)
			{
				fprintf(stderr, "M ROMTAP no good, early end-of-file in %s\nExpected 16384 bytes, got %u\n", rom, i);
				return(2);
			}
			blk[i]=c;
		}
		fclose(romf);
		if(!tapf)
		{
			fprintf(stderr, "M ROMTAP no good, failed to open output file %s\n", tap);
			perror("fopen");
			return(1);
		}
		int e;
		if((e=blk2tap(tapf, blk, rom)))
		{
			fprintf(stderr, "M ROMTAP no good, TAP generation failed with error %d\n", e);
			return(1);
		}
		fclose(tapf);
		return(0);
	}
}

int blk2tap(FILE *out, char blk[16384], char * filename)
{
	/* WRITE HEADER BLOCK */
	fputc(19, out);		// header length: 0x0013
	fputc(0, out);		// ^^
	unsigned short cksum=0;
	fputc(0, out);		// flag byte: 00=header
	fputc(3, out);		// type=CODE
	cksum^=3;
	int i;
	bool eof=false;
	for(i=0;i<10;i++)
	{
		if(eof)
		{
			fputc(0x20, out);
			cksum^=0x20;
		}
		else
		{
			if(filename[i])
			{
				fputc(filename[i], out);
				cksum^=filename[i];
			}
			else
			{
				fputc(0x20, out);
				cksum^=0x20;
				eof=true;
			}
		}
	}
	fputc(0, out);		// data length: 0x4000
	fputc(0x40, out);	// ^^
	cksum^=0x40;
	fputc(0, out);		// data address: 0xC000
	fputc(0xC0, out);	// ^^
	cksum^=0xC0;
	fputc(0, out);		// 'parameter 2': 0x8000
	fputc(0x80, out);	// ^^
	cksum^=0x80;
	fputc(cksum, out);	// checksum
	/* WRITE DATA BLOCK */
	fputc(2, out);		// data block length: 0x4002 (0x4000 data + 1 header + 1 checksum)
	fputc(0x40, out);	// ^^
	cksum=0;
	fputc(0xff, out);	// flag byte: ff=data
	cksum^=0xff;
	for(i=0;i<16384;i++)
	{
		fputc(blk[i], out);	// data byte
		cksum^=blk[i];
	}
	fputc(cksum, out);	// checksum
	return(0);
}
